// CSCE 4620 - Assignment 1 (Timers)
// Jeff Morrison
//
// NOTE: Overall OS code-structure based on Example1 from 
//         the uC/OS-II CD.

#include "includes.h"

#define  TASK_STK_SIZE                 512            /* Size of each task's stacks (# of WORDs)       */
#define  N_TASKS                        5             /* Number of tasks                               */

OS_STK        TaskStk[N_TASKS][TASK_STK_SIZE];        /* Tasks stacks                                  */
OS_STK        TaskStartStk[TASK_STK_SIZE];
INT8U         TaskData[N_TASKS];                      /* Parameters to pass to each task               */
OS_EVENT      *ChopStickMutex[N_TASKS];

        void  PhilosopherTask(void *pdata);           /* Function prototypes of philosopher-tasks      */
        void  TaskStart(void *data);                  /* Function prototypes of Startup task           */
static  void  TaskStartDispInit(void);


void main(void) {

    INT8U i;
    
    PC_DispClrScr(DISP_FGND_WHITE + DISP_BGND_BLACK);      /* Clear the screen                         */

    OSInit();                                              /* Initialize uC/OS-II                      */

    PC_DOSSaveReturn();                                    /* Save environment to return to DOS        */
    PC_VectSet(uCOS, OSCtxSw);                             /* Install uC/OS-II's context switch vector */
    
    
    OSTaskCreate(TaskStart, (void *)0, &TaskStartStk[TASK_STK_SIZE - 1], 0);
    OSStart();                                             /* Start multitasking                       */
}

void TaskStart(void *pdata) {
	
#if OS_CRITICAL_METHOD == 3                                /* Allocate storage for CPU status register */
    OS_CPU_SR  cpu_sr;
#endif
    char       s[100];
    INT16S     key;
    INT8U      i;
    INT8U err;

    pdata = pdata;                                         /* Prevent compiler warning                 */

    TaskStartDispInit();                                   /* Initialize the display                   */

    OS_ENTER_CRITICAL();
    PC_VectSet(0x08, OSTickISR);                           /* Install uC/OS-II's clock tick ISR        */
    PC_SetTickRate(OS_TICKS_PER_SEC);                      /* Reprogram tick rate                      */
    OS_EXIT_CRITICAL();

    OSStatInit();                                          /* Initialize uC/OS-II's statistics         */

    TaskStartDispInit();

    //INITIALIZE CHOPSTICK SEMAPHORES
    for (i=0; i<N_TASKS; i++)
	ChopStickMutex[i] = OSMutexCreate(i+1, &err);

    //CREATE PHILOSOPHER TASKS
    for (i=0; i<N_TASKS; i++) {
    	
    	//LET THE PHILOSOPHER KNOW HIS ID NUMBER
    	TaskData[i] = i+1;

    	//CREATE THE PHILOSOPHER
    	OSTaskCreate(PhilosopherTask, (void *)&TaskData[i], &TaskStk[i][TASK_STK_SIZE - 1], i+50);
    }

    
    for (;;) {
	
        if (PC_GetKey(&key) == TRUE) {                     /* See if key has been pressed              */
            if (key == 0x1B) {                             /* Yes, see if it's the ESCAPE key          */
                PC_DOSReturn();                            /* Return to DOS                            */
            }
        }

        OSCtxSwCtr = 0;                                    /* Clear context switch counter             */
        OSTimeDlyHMSM(0, 0, 0, 500);                         /* Wait 500 ms                          */
    }
}

static  void  TaskStartDispInit(void) {

    PC_DispStr( 0,  0, "                                Semaphores Lab 3                                ", DISP_FGND_WHITE + DISP_BGND_RED);
    PC_DispStr( 0,  1, "                                 Jeff Morrison                                  ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0,  2, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0,  3, "                    These philosophers are currently eating:                    ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0,  4, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0,  5, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0,  6, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0,  7, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0,  8, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0,  9, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0, 10, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0, 11, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0, 12, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0, 13, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0, 14, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0, 15, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0, 16, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0, 17, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0, 18, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0, 19, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0, 20, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0, 21, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0, 21, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0, 21, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
    PC_DispStr( 0, 24, "                            <-PRESS 'ESC' TO QUIT->                             ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY + DISP_BLINK);
}

static void PhilosopherTask(void *pdata) {
	
    	INT8U phil_num;
    	INT8U left_stick;
    	INT8U right_stick;
    	INT8U err;
	char philosopher_id_string[80], s[80];
	
	phil_num = *(int *)pdata;
    	left_stick  = phil_num-1;
    	right_stick = phil_num == N_TASKS ? 0 : phil_num;
	//sprintf(philosopher_id_string, "* Philosopher #%d (left:%d, right:%d)   ", phil_num, left_stick, right_stick);
	sprintf(philosopher_id_string, "* Philosopher #%d   ", phil_num, left_stick, right_stick);

	for (;;) {
		
    		//TRY TO PICK UP THE CHOPSTICK TO THE LEFT
		OSMutexPend(ChopStickMutex[left_stick], 0, &err);

		if (err != OS_NO_ERR)
			continue;//PC_DispStr(30, 15,"!",DISP_FGND_BLACK+DISP_BGND_LIGHT_GRAY);
/*
		//IF ERROR OBTAINING LEFT CHOPSTICK: START OVER AND TRY AGAIN
		switch (err) {
			
			case OS_TIMEOUT:
				sprintf(s, "PhilTask: OS_TIMEOUT (#%d, left(%d)) ", phil_num, left_stick);
				PC_DispStr(0,15,s,DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
				continue;
			case OS_ERR_EVENT_TYPE:
				PC_DispStr(0,15,"PhilTask: OS_ERR_EVENT_TYPE",DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
				continue;
			case OS_ERR_PEVENT_NULL:
				PC_DispStr(0,15,"PhilTask: OS_ERR_PEVENT_NULL",DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
				continue;
			case OS_ERR_PEND_ISR:
				PC_DispStr(0,15,"PhilTask: OS_ERR_PEND_ISR",DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
				continue;
			default:
				sprintf(s, "PhilTask: OS_NO_ERR (#%d, left(%d))", phil_num, right_stick);
				PC_DispStr(0,15,s,DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
		}
*/
		//TRY TO PICK UP THE CHOPSTICK TO THE RIGHT
		OSMutexPend(ChopStickMutex[right_stick], 0, &err);
		
		if (err != OS_NO_ERR)
			continue;//PC_DispStr(30, 15,"!",DISP_FGND_BLACK+DISP_BGND_LIGHT_GRAY);
/*
		//IF ERROR OBTAINING RIGHT CHOPSTICK: RELEASE THE LEFT CHOPSTICK, 
		//  START OVER, AND TRY AGAIN
		switch (err) {
			
			case OS_TIMEOUT:
				sprintf(s, "PhilTask: OS_TIMEOUT (#%d, right(%d))", phil_num, right_stick);
				PC_DispStr(0,15,s,DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
				continue;
			case OS_ERR_EVENT_TYPE:
				PC_DispStr(0,15,"PhilTask: OS_ERR_EVENT_TYPE",DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
				continue;
			case OS_ERR_PEVENT_NULL:
				PC_DispStr(0,15,"PhilTask: OS_ERR_PEVENT_NULL",DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
				continue;
			case OS_ERR_PEND_ISR:
				PC_DispStr(0,15,"PhilTask: OS_ERR_PEND_ISR",DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
				continue;
			default:
				sprintf(s, "PhilTask: OS_NO_ERR (#%d, right(%d))", phil_num, right_stick);
				PC_DispStr(0,15,s,DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
		}
*/
		//YELL THAT YOU'RE EATING FROM THE TOP OF A MOUNTAIN
    		PC_DispStr( 33, 4+phil_num, philosopher_id_string, DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
		
		//DELAY FOR A BIT TO EAT WITH THE CHOPSTICKS
		OSTimeDlyHMSM(0, 0, 1, 0);
		
		//PUT DOWN THE CHOPSTICKS, AND STOP YELLING
		OSMutexPost(ChopStickMutex[left_stick]);
		OSMutexPost(ChopStickMutex[right_stick]);
    		PC_DispStr( 33, 4+phil_num, "                                                                                ", DISP_FGND_BLACK + DISP_BGND_LIGHT_GRAY);
		
		//DELAY FOR A FEW TICKS TO LET SOMEONE ELSE PICK UP THE STICKS
		OSTimeDly(5);
	}
}
